#! /usr/bin/env python3

import pysat
import collections, bz2
import multiprocessing
import datetime, time
#max_years = 1  # Ref: https://github.com/python/cpython/blob/889b0b9bf95651fc05ad532cc4a66c0f8ff29fc2/Include/cpython/pytime.h
#max_sleep_time = datetime.timedelta(days=1).total_seconds()

class QCN:
    """Qualitative Constraint Network (restricted to arcs i < j)"""
    def __init__(self, signature):
        self.size = 0
        self.signature = frozenset(signature)
        s_sign = list(signature)
        s_sign.sort()
        self.universal = s_sign
        self.constraints = dict()

    def add(self, i, j, relation):
        """"Add a constraint (i < j)"""
        self.size = max(self.size, max(i, j)+1)
        assert i < j
        for base_rel in relation:
            if not base_rel in self.signature:
                raise SystemExit("Base relation not in given signature\n")

        idx = (i, j)
        old = self.signature
        if idx in self.constraints:
            old = frozenset(self.constraints[idx])
        relation_sort = list(frozenset(relation) & old)
        relation_sort.sort()

        self.constraints[idx] = relation_sort

    def get(self, i, j):
        """Read constraint on (i,j); always returns a relation."""
        try:
            return self.constraints[(i, j)]
        except KeyError:
            return self.universal

    def is_2_consistent(self):
        """Test 2-consistency"""
        for i, j in self.iterate_strict_triangle():
            if not self.get(i, j):
                return False
        return True

    def iterate_strict_triangle(self):
        """Iterate all pairs i < j on current primal constraint graph."""
        for i in range(0, self.size):
            for j in range(i+1, self.size):
                yield i, j


    def iterate_strict_triples(self):
        """iterate i < j < k triples"""
        for i in range(0, self.size):
            for j in range(i+1, self.size):
                for k in range(j+1, self.size):
                    yield i, j, k


class PropositionalAtoms: 
    """Dictionary enumerating variable strings."""
    def __init__(self):
        self.names = dict()
        self.aux = dict()
        self.last_used = 0
    
    def encode(self, i, j, baserel, d):
        try:
            return self.names[i][j][baserel][d]
        except KeyError:
            assert i <= j
            self.last_used += 1

            if not i in self.names:
                self.names[i] = dict()
            if not j in self.names[i]:
                self.names[i][j] = dict()
            if not baserel in self.names[i][j]:
                self.names[i][j][baserel] = dict()
            self.names[i][j][baserel][d] = self.last_used
            return self.last_used            
            
    def encode_aux(self, d):
        try:
            return self.aux[d]
        except KeyError:
            self.last_used += 1
            
            self.aux[d] = self.last_used
            return self.last_used

def read_comp_table(calculus):
    """Read an GQR-format composition table"""
    import re

    table = { }
    all_relations = set()

    lines = open(calculus, 'r')
    for line in lines:
        extract = re.search('^(.*):(.*)::[ ]*\((.*)\)$', line)
        assert(extract)
        left = extract.group(1).strip()
        right = extract.group(2).strip()
        supp = extract.group(3).strip()
        table[left + " " + right] = frozenset(supp.split())
        all_relations.add(left)
        all_relations.add(right)
    lines.close()

    return table, frozenset(all_relations)

def read_gqr_csp(buffer, signature):
    """Read an GQR-format CSP from input"""
    import re

    data = QCN(signature)
    
    ID = buffer.pop(0).split()[1].strip()

    for line in buffer:
        res = re.search('^[ ]*([0-9]*)[ ]*([0-9]*) \((.*)\)', line)
        if res:
            i = int(res.group(1))
            j = int(res.group(2))
            assert 0 <= i < j
            rel = res.group(3).strip().split(" ")
            if not rel:
                return None
            for base_rel in rel:
                if not base_rel in signature:
                    raise SystemExit("Calculus signature does not match CSP")

            if frozenset(rel) != signature:  # ignore universal statements
                data.add(i, j, rel)

    return data, ID


def dom_encoding(qcn, atoms, clauses, wclauses, l):

    for i, j in qcn.iterate_strict_triangle():
    
        alod_clause = []
        
        for d in range(1,l+1):
        
            rel = qcn.get(i, j)
            alod = [ atoms.encode(i, j, br, d) for br in rel ]
            alod_clause.extend(alod)
            
        clauses.append(alod_clause)

    for d in range(1,l+1):
        for i, j in qcn.iterate_strict_triangle():

            rel = qcn.universal
            alo = [ atoms.encode(i, j, br, d) for br in rel ]

            clauses.append(alo)
        
            for idx1 in range(len(rel)):
                br1 = rel[idx1]
                for idx2 in range(idx1+1, len(rel)):
                    br2 = rel[idx2]
                    amo = [ -1 * atoms.encode(i, j, br1, d),
                        -1 * atoms.encode(i, j, br2, d) ]
                    clauses.append(amo)
                                    
    
def encoding(qcn, comptable, l):
    clauses = []
    wclauses = []
    
    atoms = PropositionalAtoms()

    dom_encoding(qcn, atoms, clauses, wclauses, l)

    for i, j, k in qcn.iterate_strict_triples():
        r_ij = qcn.universal
        r_ik = qcn.universal
        r_jk = qcn.universal
        
        for d in range(1,l+1):
        
            for br1 in r_ij:
                not_b_ij = -1 * atoms.encode(i, j, br1, d)
                for br2 in r_jk:
                    intersection = list(frozenset(r_ik) & comptable[br1 +" "+ br2])
                    intersection.sort()

                    clause = [ not_b_ij, -1 * atoms.encode(j, k, br2, d) ]
                    clause += [ atoms.encode(i, k, br, d) for br in intersection ]
                    clauses.append(clause)  
                    
    for d in range(1,l+1):
    
        aux_var = atoms.encode_aux(d)      
        for i, j in qcn.iterate_strict_triangle():
        
            rel = qcn.get(i, j)
            compl_rel = sorted((qcn.signature - set(rel)))
            compl = [ atoms.encode(i, j, br, d) for br in compl_rel ]
            
            for b in compl:
                clauses.append([-1*b, aux_var])
        wclauses.append(([-1*aux_var],1))

    return atoms, clauses, wclauses

def check_options():
    """Parse options with assertions"""
    from argparse import ArgumentParser

    parser = ArgumentParser()

    parser.add_argument('GQR_COMPOSITION_TABLE_FILE')
    parser.add_argument("-f", "--file", action="store", required=True, type=str, dest="filename")
    parser.add_argument("-t", "--timeout", action="store", required=False, type=int, dest="timeout")

    result = parser.parse_args()

    return result.filename, result.timeout,\
        result.GQR_COMPOSITION_TABLE_FILE
        
def foo(ID, INPUT_QCN, COMPTABLE):

    l = 1
    total_t = 0
             
    while True:

        atoms, clauses, wclauses = encoding(INPUT_QCN, COMPTABLE, l)

        wcnf = WCNF()            

        for clause in clauses:
            wcnf.append(clause)
       
        for wclause in wclauses:
            wcnf.append(wclause[0], weight=wclause[1])               

        with RC2S(wcnf) as rc2s:
            model = rc2s.compute()
            total_t += rc2s.oracle_time()
            if model:
               cleanModel = set([i for i in model if i>0])
               allCons = len(INPUT_QCN.constraints)
               commonCons = 0
               for (i,j) in INPUT_QCN.constraints:
                   flag = True
                   cc = [0]*l
                   for d in range(1,l+1):
                       for b in INPUT_QCN.constraints[(i,j)]:
                           if atoms.encode(i, j, b, d) in cleanModel:
                              cc[d-1] = 1
                              break
                   if sum(cc) == l: commonCons += 1
               print("QCN:", ID, rc2s.cost, total_t, "with", l, "components", "and", commonCons, "/", allCons, "=", commonCons/float(allCons), "similarity")
               rc2s.delete()           
               break
            rc2s.delete()            
        l += 1

if __name__ == '__main__':
    QCN_FILENAME, TIMEOUT, CT_FILENAME = check_options()

    COMPTABLE, SIGNATURE = read_comp_table(CT_FILENAME)
    
    from pysat.examples.rc2 import RC2 as RC2S
    from pysat.formula import WCNF
    
    with open(QCN_FILENAME) as f:
       
       total = 0
       meter = 0
       buffer = []

       for i in f:
          buffer.append(i)

          if i.strip() == '.':
             INPUT_QCN, ID = read_gqr_csp(buffer, SIGNATURE)
             buffer.clear()

             if INPUT_QCN.size == 0 or not INPUT_QCN.is_2_consistent():
                # no constraints read (assume problem was unsatisfiable) or
                # problem is not 2-consistency (=> unsat)
                print("p cnf 1 2")
                print("1 0")
                print("-1 0")
                raise SystemExit()
                
             p = multiprocessing.Process(target=foo, name="Foo", args=(ID, INPUT_QCN, COMPTABLE))
             p.start()
             p.join(TIMEOUT)
             
             if p.is_alive():
                print("QCN:", ID, "timeout") 

                # Terminate foo
                p.terminate()
                p.join()
             
